/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/ecc/ecc.c                                                *
 * Created:     2003-06-20 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2003-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include "hamming.h"
#include "ecc.h"


static const char *argv0 = NULL;

static int      par_encode = 0;
static int      par_decode = 0;
static int      par_verbose = 1;
static unsigned par_c = 0;
static unsigned par_n = 11;
static unsigned par_k = 4;
static unsigned par_d = 4096;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'c', 1, "ecc-file", "string", "Set the ECC file name [stdout]" },
	{ 'd', 0, "decode", NULL, "Decode a stream" },
	{ 'e', 0, "encode", NULL, "Encode a stream" },
	{ 'f', 1, "distance", "int", "Set the byte distance [4096]" },
	{ 'i', 1, "input", "string", "Set input file name [stdin]" },
	{ 'I', 1, "info", "string", "Print ecc file info" },
	{ 'k', 1, "parity-bits", "int", "Set the number of parity bits [4]" },
	{ 'n', 1, "data-bits", "int", "Set the number of data bits [11]" },
	{ 'N', 1, "code-bits", "int", "Set the number of code bits [0]" },
	{ 'o', 1, "output", "string", "Set output file name [none]" },
	{ 'q', 0, "quiet", NULL, "Quiet operation [no]" },
	{ 'v', 0, "verbose", NULL, "Verbose operation [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"ecc: calculate error correction codes",
		"usage: ecc [options] [input [ecc [output]]]",
		opts
	);

	fflush (stdout);
}

static
void ecc_set_uint32 (void *buf, unsigned long i, unsigned long val)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (val >> 24) & 0xff;
	tmp[1] = (val >> 16) & 0xff;
	tmp[2] = (val >> 8) & 0xff;
	tmp[3] = val & 0xff;
}

static
void ecc_set_uint16 (void *buf, unsigned long i, unsigned short val)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (val >> 8) & 0xff;
	tmp[1] = val & 0xff;
}

static
unsigned long ecc_get_uint32 (const void *buf, unsigned long i)
{
	unsigned long       ret;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	ret = tmp[0] & 0xff;
	ret = (ret << 8) | (tmp[1] & 0xff);
	ret = (ret << 8) | (tmp[2] & 0xff);
	ret = (ret << 8) | (tmp[3] & 0xff);

	return (ret);
}

static
unsigned short ecc_get_uint16 (const void *buf, unsigned long i)
{
	unsigned short      ret;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	ret = ((tmp[0] & 0xff) << 8) | (tmp[1] & 0xff);

	return (ret);
}

static
buffer_t *buf_alloc (unsigned n, unsigned d)
{
	unsigned i;
	buffer_t *ret;

	ret = malloc (sizeof (buffer_t));
	if (ret == NULL) {
		return (NULL);
	}

	ret->n = n;
	ret->d = d;
	ret->bytes = 0;

	ret->buf = malloc (n * sizeof (unsigned char *));
	if (ret->buf == NULL) {
		free (ret);
		return (NULL);
	}

	ret->buf[0] = malloc ((unsigned long) n * (unsigned long) d);
	if (ret->buf[0] == NULL) {
		free (ret->buf);
		free (ret);
		return (NULL);
	}

	for (i = 1; i < n; i++) {
		ret->buf[i] = ret->buf[i - 1] + d;
	}

	return (ret);
}

static
void buf_free (buffer_t *buf)
{
	if (buf != NULL) {
		free (buf->buf[0]);
		free (buf->buf);
		free (buf);
	}
}

static
void buf_clear (buffer_t *buf, unsigned long bytes)
{
	unsigned long max;

	max = (unsigned long) buf->n * (unsigned long) buf->d;

	if (bytes > max) {
		bytes = max;
	}

	memset (buf->buf[0], 0, max);

	buf->bytes = bytes;
}

static
unsigned buf_read (buffer_t *buf, FILE *fp)
{
	unsigned long cnt;

	cnt = (unsigned long) buf->n * (unsigned long) buf->d;

	buf->bytes = fread (buf->buf[0], 1, cnt, fp);

	if (buf->bytes < cnt) {
		memset (buf->buf[0] + buf->bytes, 0, cnt - buf->bytes);
	}

	if (buf->bytes < buf->d) {
		return (buf->bytes);
	}

	return (buf->d);
}

static
int buf_write (buffer_t *buf, FILE *fp)
{
	if (fwrite (buf->buf[0], 1, buf->bytes, fp) != buf->bytes) {
		return (1);
	}

	return (0);
}

static
void buf_get_words (buffer_t *buf, unsigned idx, unsigned char *ret)
{
	unsigned i;

	if (idx >= buf->d) {
		return;
	}

	for (i = 0; i < buf->n; i++) {
		ret[i] = buf->buf[i][idx];
	}
}

static
void buf_set_words (buffer_t *buf, unsigned idx, const unsigned char *val)
{
	unsigned i;

	if (idx >= buf->d) {
		return;
	}

	for (i = 0; i < buf->n; i++) {
		buf->buf[i][idx] = val[i];
	}
}

static
ecc_t *ecc_new (unsigned code, unsigned n, unsigned k, unsigned d)
{
	ecc_t *ret;

	ret = malloc (sizeof (ecc_t));
	if (ret == NULL) {
		return (NULL);
	}

	ret->code = code;
	ret->n = n;
	ret->k = k;
	ret->d = d;

	ret->tmpdat = malloc (n + k + k);
	if (ret->tmpdat == NULL) {
		free (ret);
		return (NULL);
	}

	ret->tmpecc1 = ret->tmpdat + n;
	ret->tmpecc2 = ret->tmpdat + n + k;

	ret->bufdat = buf_alloc (n, d);
	ret->bufecc = buf_alloc (k, d);

	if ((ret->bufdat == NULL) || (ret->bufecc == NULL)) {
		buf_free (ret->bufdat);
		buf_free (ret->bufecc);
		free (ret->tmpdat);
		free (ret);
		return (NULL);
	}

	return (ret);
}

static
void ecc_del (ecc_t *ecc)
{
	if (ecc != NULL) {
		buf_free (ecc->bufecc);
		buf_free (ecc->bufdat);
		free (ecc->tmpdat);
		free (ecc);
	}
}

static
const char *ecc_get_code_string (unsigned code)
{
	switch (code) {
	case ECC_CODE_HAMMING:
		return ("hamming");
	}

	return ("<unknown>");
}

static
void ecc_print_info (const ecc_t *ecc, FILE *fp)
{
	const char *code;

	code = ecc_get_code_string (ecc->code);

	fprintf (fp, "code type:      0x%08lx (%s)\n", ecc->code, code);
	fprintf (fp, "data bits:      %u\n", ecc->n);
	fprintf (fp, "check bits:     %u\n", ecc->k);
	fprintf (fp, "offset:         %u\n", ecc->d);
	fprintf (fp, "code:           %s (%u, %u, %u)\n",
		code, ecc->n + ecc->k, ecc->k, ecc->d
	);
	fprintf (fp, "code rate:      %.2f%%\n",
		100.0 * (double) ecc->k / (double) (ecc->n + ecc->k)
	);
	fprintf (fp, "overhead:       %.2f%%\n",
		100.0 * (double) ecc->k / (double) ecc->n
	);
}

static
void ecc_print_header_info (const ecc_header_t *hdr, FILE *fp)
{
	const char *code;

	code = ecc_get_code_string (hdr->code);

	fprintf (fp, "magic:          0x%08lx (%c%c%c%c)\n",
		hdr->magic,
		(int) ((hdr->magic >> 24) & 0xff),
		(int) ((hdr->magic >> 16) & 0xff),
		(int) ((hdr->magic >> 8) & 0xff),
		(int) (hdr->magic & 0xff)
	);

	fprintf (fp, "format version: %lu\n", hdr->vers);
	fprintf (fp, "header size:    %lu\n", hdr->size);
	fprintf (fp, "code type:      0x%08lx (%s)\n", hdr->code, code);
	fprintf (fp, "data bits:      %u\n", hdr->n);
	fprintf (fp, "check bits:     %u\n", hdr->k);
	fprintf (fp, "offset:         %lu\n", hdr->d);
	fprintf (fp, "code:           %s (%u, %u, %lu)\n",
		code, hdr->n + hdr->k, hdr->k, hdr->d
	);
	fprintf (fp, "code rate:      %.2f%%\n",
		100.0 * (double) hdr->k / (double) (hdr->n + hdr->k)
	);
	fprintf (fp, "overhead:       %.2f%%\n",
		100.0 * (double) hdr->k / (double) hdr->n
	);
}

static
int ecc_write_header (ecc_t *ecc, FILE *fp)
{
	unsigned char buf[32];

	ecc_set_uint32 (buf, 0, ECC_MAGIC);
	ecc_set_uint32 (buf, 4, 24);
	ecc_set_uint32 (buf, 8, 0);
	ecc_set_uint32 (buf, 12, ECC_CODE_HAMMING);
	ecc_set_uint16 (buf, 16, ecc->n);
	ecc_set_uint16 (buf, 18, ecc->k);
	ecc_set_uint32 (buf, 20, ecc->d);

	if (fwrite (buf, 1, 24, fp) != 24) {
		return (1);
	}

	return (0);
}

static
int ecc_read_header (FILE *fp, ecc_header_t *hdr)
{
	unsigned char buf[32];

	if (fread (buf, 1, 8, fp) != 8) {
		fprintf (stderr, "%s: file read error\n", argv0);
		return (1);
	}

	hdr->magic = ecc_get_uint32 (buf, 0);
	if (hdr->magic != ECC_MAGIC) {
		fprintf (stderr, "%s: not an ECC file (bad magic)\n", argv0);
		return (1);
	}

	hdr->size = ecc_get_uint32 (buf, 4);

	if (hdr->size == 0) {
		hdr->vers = 0;
		hdr->size = 16;
		hdr->code = ECC_CODE_HAMMING;

		if (fread (buf + 8, 1, 8, fp) != 8) {
			fprintf (stderr, "%s: file read error\n", argv0);
			return (1);
		}

		hdr->n = ecc_get_uint16 (buf, 8);
		hdr->k = ecc_get_uint16 (buf, 10);
		hdr->d = ecc_get_uint32 (buf, 12);
	}
	else {
		if (hdr->size < 24) {
			fprintf (stderr, "%s: header too small (%lu bytes)\n",
				argv0, hdr->size
			);
			return (1);
		}

		if (fread (buf + 8, 1, 16, fp) != 16) {
			fprintf (stderr, "%s: file read error\n", argv0);
			return (1);
		}

		hdr->vers = ecc_get_uint32 (buf, 8);
		hdr->code = ecc_get_uint32 (buf, 12);
		hdr->n = ecc_get_uint16 (buf, 16);
		hdr->k = ecc_get_uint16 (buf, 18);
		hdr->d = ecc_get_uint32 (buf, 20);

		if (hdr->size > 24) {
			unsigned long n = hdr->size - 24;

			while (n > 0) {
				fgetc (fp);
				n -= 1;
			}
		}
	}

	return (0);
}

static
ecc_t *ecc_open (FILE *fp)
{
	ecc_t        *ecc;
	ecc_header_t hdr;

	if (ecc_read_header (fp, &hdr)) {
		return (NULL);
	}

	ecc = ecc_new (hdr.code, hdr.n, hdr.k, hdr.d);

	return (ecc);
}

static
int ecc_encode (FILE *finp, FILE *fout, FILE *fecc)
{
	unsigned i, n;
	ecc_t    *ecc;

	ecc = ecc_new (ECC_CODE_HAMMING, par_n, par_k, par_d);
	if (ecc == NULL) {
		return (1);
	}

	if (par_verbose > 1) {
		ecc_print_info (ecc, stderr);
	}

	if (ecc_write_header (ecc, fecc)) {
		ecc_del (ecc);
		return (1);
	}

	n = buf_read (ecc->bufdat, finp);

	while (n > 0) {
		buf_clear (ecc->bufecc, (unsigned long) ecc->d * ecc->k);

		for (i = 0; i < n; i++) {
			buf_get_words (ecc->bufdat, i, ecc->tmpdat);
			hamming_calc (ecc->n, ecc->tmpdat, ecc->k, ecc->tmpecc1);
			buf_set_words (ecc->bufecc, i, ecc->tmpecc1);
		}

		if (fout != NULL) {
			if (buf_write (ecc->bufdat, fout)) {
				fprintf (stderr, "%s: write error (out)\n",
					argv0
				);
				ecc_del (ecc);
				return (1);
			}
		}

		if (buf_write (ecc->bufecc, fecc)) {
			fprintf (stderr, "%s: write error (ecc)\n", argv0);
			ecc_del (ecc);
			return (1);
		}

		n = buf_read (ecc->bufdat, finp);
	}

	ecc_del (ecc);

	return (0);
}

static
int ecc_decode (FILE *finp, FILE *fout, FILE *fecc)
{
	unsigned      i, n;
	unsigned long cnt;
	ecc_t         *ecc;

	ecc = ecc_open (fecc);
	if (ecc == NULL) {
		return (1);
	}

	if (par_verbose > 1) {
		ecc_print_info (ecc, stderr);
	}

	if (ecc->code != ECC_CODE_HAMMING) {
		fprintf (stderr, "%s: unknown code type (0x%08lx)\n",
			argv0, ecc->code
		);
		ecc_del (ecc);
		return (1);
	}

	cnt = 0;

	n = buf_read (ecc->bufdat, finp);
	if (buf_read (ecc->bufecc, fecc) < n) {
		ecc_del (ecc);
		return (1);
	}

	while (n > 0) {
		for (i = 0; i < n; i++) {
			buf_get_words (ecc->bufdat, i, ecc->tmpdat);
			buf_get_words (ecc->bufecc, i, ecc->tmpecc1);

			hamming_calc (ecc->n, ecc->tmpdat, ecc->k, ecc->tmpecc2);
			cnt += hamming_correct (ecc->n, ecc->tmpdat, ecc->k, ecc->tmpecc1, ecc->tmpecc2);

			buf_set_words (ecc->bufdat, i, ecc->tmpdat);
		}

		if (fout != NULL) {
			if (buf_write (ecc->bufdat, fout)) {
				fprintf (stderr, "%s: write error\n", argv0);
				ecc_del (ecc);
				return (1);
			}
		}

		n = buf_read (ecc->bufdat, finp);
		if (buf_read (ecc->bufecc, fecc) < n) {
			fprintf (stderr, "%s: read error\n", argv0);
			ecc_del (ecc);
			return (1);
		}
	}

	ecc_del (ecc);

	if (par_verbose > 0) {
		if (cnt > 0) {
			fprintf (stderr, "%s: corrected %lu bits\n",
				argv0, cnt
			);
		}
	}

	return (0);
}

static
int ecc_info (const char *fname)
{
	int          r;
	FILE         *fp;
	ecc_header_t hdr;

	fp = misc_open_inp (argv0, fname, "rb");

	if (fp == NULL) {
		return (1);
	}

	r = ecc_read_header (fp, &hdr);

	misc_close (fp);

	if (r) {
		fprintf (stderr, "reading ECC header failed\n");
		return (1);
	}

	ecc_print_header_info (&hdr, stdout);

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *ninp, *nout, *necc;
	FILE       *finp, *fout, *fecc;

	argv0 = argv[0];

	ninp = NULL;
	nout = NULL;
	necc = NULL;

	finp = NULL;
	fout = NULL;
	fecc = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("ecc");
			return (EXIT_SUCCESS);

		case 'c':
			necc = optarg[0];
			break;

		case 'd':
			par_encode = 0;
			par_decode = 1;
			break;

		case 'e':
			par_encode = 1;
			par_decode = 0;
			break;

		case 'f':
			if (str_get_uint (optarg[0], &par_d)) {
				fprintf (stderr,
					"%s: bad byte distance (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'i':
			ninp = optarg[0];
			break;

		case 'I':
			if (ecc_info (optarg[0])) {
				return (EXIT_FAILURE);
			}
			break;

		case 'k':
			if (str_get_uint (optarg[0], &par_k)) {
				fprintf (stderr,
					"%s: bad number of parity bits (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'n':
			if (str_get_uint (optarg[0], &par_n)) {
				fprintf (stderr,
					"%s: bad number of data bits (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'N':
			if (str_get_uint (optarg[0], &par_c)) {
				fprintf (stderr,
					"%s: bad number of code bits (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			nout = optarg[0];
			break;

		case 'q':
			par_verbose -= 1;
			break;

		case 'v':
			par_verbose += 1;
			break;

		case 0:
			if (ninp == NULL) {
				ninp = optarg[0];
			}
			else if (necc == NULL) {
				necc = optarg[0];
			}
			else if (nout == NULL) {
				nout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if ((par_encode == 0) && (par_decode == 0)) {
		return (EXIT_SUCCESS);
	}

	if (par_d == 0) {
		fprintf (stderr, "%s: byte distance can't be zero\n", argv0);
		return (EXIT_FAILURE);
	}

	if (par_k > 16) {
		fprintf (stderr, "%s: too many parity bits (16 max)\n", argv0);
		return (EXIT_FAILURE);
	}

	if (par_k == 0) {
		fprintf (stderr, "%s: check bit count can't be zero\n", argv0);
		return (EXIT_FAILURE);
	}

	if (par_c > 0) {
		if (par_c <= par_k) {
			fprintf (stderr, "%s: (%u, %u) is not a valid code\n",
				argv0, par_c, par_k
			);
			return (EXIT_FAILURE);
		}

		par_n = par_c - par_k;
	}

	if (par_n == 0) {
		if (par_k == 1) {
			par_n = 1;
		}
		else {
			par_n = (1 << par_k) - par_k - 1;
		}
	}

	par_c = par_n + par_k;

	if (par_verbose) {
		if ((par_n + par_k) >= (1 << par_k)) {
			fprintf (stderr,
				"%s: warning: (%u, %u) is not a legal hamming code\n",
				argv0, par_c, par_k
			);
		}
	}

	finp = misc_open_inp (argv0, ninp, "rb");

	if (finp == NULL) {
		return (EXIT_FAILURE);
	}

	if (nout != NULL) {
		fout = misc_open_out (argv0, nout, "wb");

		if (fout == NULL) {
			return (EXIT_FAILURE);
		}
	}

	if (par_encode) {
		fecc = misc_open_out (argv0, necc, "wb");

		if (fecc == NULL) {
			return (EXIT_FAILURE);
		}

		r = ecc_encode (finp, fout, fecc);
	}
	else {
		fecc = misc_open_inp (argv0, necc, "rb");

		if (fecc == NULL) {
			return (EXIT_FAILURE);
		}

		r = ecc_decode (finp, fout, fecc);
	}

	misc_close (fecc);
	misc_close (fout);
	misc_close (finp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
