/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/entab/entab.c                                            *
 * Created:     2000-12-10 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include "entab.h"


static const char *argv0 = NULL;

static int      par_entab = 1;
static int      par_trim = 0;
static int      par_beg = 0;
static unsigned par_min_step = 0;

static tab_lst  *par_tablst = NULL;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'b', 0, "begin", NULL, "Restrict modifications to beginning of lines [no]" },
	{ 'd', 0, "detab", NULL, "Replace tabs by spaces [no]" },
	{ 'e', 0, "entab", NULL, "Replace spaces by tabs [default]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'm', 1, "min-step", "int", "Minimum number of spaces replaceable by a tab [1]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'r', 0, "trim", NULL, "Remove spaces and tabs from end of lines [no]" },
	{ 's', 1, "default-tab", "int", "Set the default tabulator width [8]" },
	{ 't', 1, "tab", "[+]int", "Set a (relative) tab" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"entab: convert spaces to tabs and vice versa",
		"usage: entab [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
int par_get_tab (const char *str, unsigned *val)
{
	unsigned        i;
	int             rel;
	static unsigned last = 0;

	rel = 0;

	i = 0;
	while (str[i] == ' ') {
		i += 1;
	}

	if (str[i] == '+') {
		rel = 1;
		i += 1;
	}

	if (str_get_uint (str + i, val)) {
		return (1);
	}

	if (rel) {
		*val += last;
	}

	last = *val;

	return (0);
}

static
line_buf *buf_new (unsigned n)
{
	line_buf *buf;

	if (n == 0) {
		n = 1;
	}

	buf = malloc (sizeof (line_buf));

	if (buf == NULL) {
		return (NULL);
	}

	buf->chrcnt = 0;
	buf->chrmax = n;
	buf->chr = malloc (n);

	if (buf->chr == NULL) {
		free (buf);
		return (NULL);
	}

	buf->eolcnt = 0;

	return (buf);
}

static
void buf_free (line_buf *buf)
{
	if (buf != NULL) {
		free (buf->chr);
	}

	free (buf);
}

static
int buf_set_size (line_buf *buf, unsigned n)
{
	if (n < buf->chrcnt) {
		n = buf->chrcnt;
	}

	if (n == buf->chrmax) {
		return (0);
	}

	buf->chr = realloc (buf->chr, n);

	if (buf->chr == NULL) {
		return (1);
	}

	buf->chrmax = n;

	return (0);
}

static
void buf_add_char (line_buf *buf, char c)
{
	unsigned n;

	if (buf->chrcnt >= buf->chrmax) {
		n = 2 * buf->chrmax;

		if (n <= buf->chrcnt) {
			return;
		}

		buf_set_size (buf, n);
	}

	if (buf->chrcnt < buf->chrmax) {
		buf->chr[buf->chrcnt] = c;
		buf->chrcnt++;
	}
}

static
void buf_clear (line_buf *buf)
{
	buf->chrcnt = 0;
	buf->eolcnt = 0;

	if (buf->chrmax > 1024) {
		buf_set_size (buf, 1024);
	}
}

static
tab_lst *tab_new (void)
{
	tab_lst *ret;

	ret = malloc (sizeof (tab_lst));

	if (ret == NULL) {
		return (NULL);
	}

	ret->deftab = 8;
	ret->tab = NULL;

	return (ret);
}

static
void tab_free (tab_lst *lst)
{
	tab_lst_el *tab1, *tab2;

	if (lst != NULL) {
		tab1 = lst->tab;

		while (tab1 != NULL) {
			tab2 = tab1;
			tab1 = tab1->next;
			free (tab2);
		}

		free (lst);
	}
}

static
void tab_add (tab_lst *lst, unsigned pos)
{
	tab_lst_el *tab, *tmp;

	tab = malloc (sizeof (tab_lst_el));

	if (tab == NULL) {
		return;
	}

	tab->next = NULL;
	tab->pos = pos;

	if (lst->tab == NULL) {
		lst->tab = tab;
		return;
	}

	if (tab->pos < lst->tab->pos) {
		tab->next = lst->tab;
		lst->tab = tab;
		return;
	}

	tmp = lst->tab;
	while ((tmp->next != NULL) && (tmp->next->pos < tab->pos)) {
		tmp = tmp->next;
	}

	tab->next = tmp->next;
	tmp->next = tab;
}

static
unsigned tab_get_pos (tab_lst *lst, unsigned pos)
{
	tab_lst_el *tab;

	tab = lst->tab;
	while ((tab != NULL) && (tab->pos < pos)) {
		tab = tab->next;
	}

	if (tab != NULL) {
		return (tab->pos);
	}

	pos = pos + lst->deftab - 1;
	pos = pos / lst->deftab;
	pos = pos * lst->deftab;

	return (pos);
}

static
int ln_read (line_buf *buf, FILE *inp)
{
	int c;

	buf_clear (buf);

	c = fgetc (inp);
	if (c == EOF) {
		return (1);
	}

	while (c != EOF) {
		if (c == 0x0d) {
			buf->eol[0] = 0x0d;
			buf->eolcnt = 1;

			c = fgetc (inp);

			if (c != 0x0a) {
				if (c != EOF) {
					ungetc (c, inp);
				}
			}
			else {
				buf->eol[1] = 0x0a;
				buf->eolcnt = 2;
			}

			return (0);
		}
		else if (c == 0x0a) {
			buf->eol[0] = 0x0a;
			buf->eolcnt = 1;

			return (0);
		}
		else {
			buf_add_char (buf, c);

			c = fgetc (inp);
		}
	}

	return (0);
}

static
int ln_write (line_buf *buf, FILE *out)
{
	if (fwrite (buf->chr, 1, buf->chrcnt, out) != buf->chrcnt) {
		return (1);
	}

	if (fwrite (buf->eol, 1, buf->eolcnt, out) != buf->eolcnt) {
		return (1);
	}

	return (0);
}

static
void ln_rtrim (line_buf *buf)
{
	unsigned i;

	i = buf->chrcnt;

	while (i > 0) {
		i--;

		if ((buf->chr[i] == 0x20) || (buf->chr[i] == 0x09)) {
			buf->chrcnt--;
		}
		else {
			return;
		}
	}
}

static
void ln_entab (tab_lst *lst, line_buf *b_in, line_buf *b_out)
{
	unsigned i;
	char     c;
	unsigned col1, col2;
	unsigned tab;

	buf_clear (b_out);

	b_out->eolcnt = b_in->eolcnt;

	for (i = 0; i < b_in->eolcnt; i++) {
		b_out->eol[i] = b_in->eol[i];
	}

	col1 = b_out->chrcnt;
	col2 = b_out->chrcnt;

	for (i = 0; i < b_in->chrcnt; i++) {
		c = b_in->chr[i];

		switch (c) {
		case 0x20:
			col2++;
			break;

		case 0x09:
			col2 = tab_get_pos (lst, col2 + 1);
			break;

		default:
			if (col1 < col2) {
				tab = tab_get_pos (lst, col1 + 1);

				while ((tab <= col2) && ((col1 + par_min_step) < col2)) {
					buf_add_char (b_out, 0x09);
					col1 = tab;
					tab = tab_get_pos (lst, col1 + 1);
				}

				while (col1 < col2) {
					buf_add_char (b_out, 0x20);
					col1++;
				}
			}

			buf_add_char (b_out, (char) c);
			col1++;
			col2++;

			if (par_beg) {
				i++;

				while (i < b_in->chrcnt) {
					buf_add_char (b_out, b_in->chr[i]);
					i++;
				}

				return;
			}
			break;
		}
	}

	if (col1 < col2) {
		tab = tab_get_pos (lst, col1 + 1);

		while (tab <= col2) {
			buf_add_char (b_out, 0x09);
			col1 = tab;
			tab = tab_get_pos (lst, col1 + 1);
		}

		while (col1 < col2) {
			 buf_add_char (b_out, 0x20);
			 col1++;
		}
	}
}

static
void ln_detab (tab_lst *lst, line_buf *b_in, line_buf *b_out)
{
	unsigned i;
	char     c;
	unsigned col;
	unsigned tab;

	buf_clear (b_out);

	b_out->eolcnt = b_in->eolcnt;

	for (i = 0; i < b_in->eolcnt; i++) {
		b_out->eol[i] = b_in->eol[i];
	}

	col = b_out->chrcnt;

	for (i = 0; i < b_in->chrcnt; i++) {
		c = b_in->chr[i];

		switch (c) {
		case 0x09:
			tab = tab_get_pos (lst, col + 1 + par_min_step);

			while (col < tab) {
				buf_add_char (b_out, 0x20);
				col++;
			}
			break;

		case 0x20:
			buf_add_char (b_out, c);
			col++;
			break;

		default:
			buf_add_char (b_out, c);
			col++;

			if (par_beg) {
				i++;

				while (i < b_in->chrcnt) {
					buf_add_char (b_out, b_in->chr[i]);
					i++;
				}

				return;
			}
			break;
		}
	}
}

static
int entab (FILE *inp, FILE *out)
{
	int      r;
	line_buf *buf1, *buf2, *tmp;

	buf1 = buf_new (128);
	buf2 = buf_new (128);

	r = ln_read (buf1, inp);

	while (r == 0) {
		if (par_trim) {
			ln_rtrim (buf1);
		}

		if (par_entab) {
			ln_entab (par_tablst, buf1, buf2);
		}
		else {
			ln_detab (par_tablst, buf1, buf2);
		}

		tmp = buf1;
		buf1 = buf2;
		buf2 = tmp;

		if (ln_write (buf1, out)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		r = ln_read (buf1, inp);
	}

	buf_free (buf2);
	buf_free (buf1);

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	unsigned   tab;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	par_tablst = tab_new();
	if (par_tablst == NULL) {
		return (EXIT_FAILURE);
	}

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("entab");
			return (EXIT_SUCCESS);

		case 'b':
			par_beg = 1;
			break;

		case 'd':
			par_entab = 0;
			break;

		case 'e':
			par_entab = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'm':
			if (str_get_uint (optarg[0], &par_min_step)) {
				fprintf (stderr,
					"%s: bad minimum step size (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			if (par_min_step > 0) {
				par_min_step -= 1;
			}
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'r':
			par_trim = 1;
			break;

		case 's':
			if (str_get_uint (optarg[0], &par_tablst->deftab)) {
				fprintf (stderr,
					"%s: bad default tab width (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			if (par_tablst->deftab == 0) {
				fprintf (stderr, "%s: tab width can't be 0\n",
					argv0
				);
				return (EXIT_FAILURE);
			}
			break;

		case 't':
			if (par_get_tab (optarg[0], &tab)) {
				fprintf (stderr, "%s: bad tab position (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			tab_add (par_tablst, tab);
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, "r");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, "w");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	r = entab (inp, out);

	misc_close (out);
	misc_close (inp);

	tab_free (par_tablst);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
