/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/evalf/evalf.c                                            *
 * Created:     2007-01-19 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2007-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/mt19937.h>
#include <lib/strtoint.h>

#include <math.h>
#include <time.h>

#ifdef HAVE_GETTIMEOFDAY
#include <sys/time.h>
#endif

#include "evalf.h"


#ifdef MISC_DOS
typedef unsigned long uint64_t;
#endif


static int evalf_literal (const char **str, double *val);
static int evalf_expr (const char **str, double *val);


static const char *argv0 = NULL;

static int        par_format_sign = 0;
static int        par_format_zero = 0;
static unsigned   par_format_width = 0;
static unsigned   par_format_precision = 0;
static const char *par_format_mode = "f";

static int        par_mt_init = 0;
static mt19937_t  par_mt;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'e', 1, "expression", "expr", "Evaluate an expression" },
	{ 'f', 1, "format", "string", "Set the format string [0.18g]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"evalf: evaluate floating point expressions",
		"usage: evalf [options] [expression...]",
		opts
	);

	fflush (stdout);
}

#ifdef MISC_NEED_TRUNC

#define trunc(x) misc_trunc(x)

static
double misc_trunc (double x)
{
	if (x < 0.0) {
		return (ceil (x));
	}
	else {
		return (floor (x));
	}
}

#endif

#ifdef MISC_NEED_ROUND

#define round(x) misc_round(x)

static
double misc_round (double x)
{
	if (x < 0.0) {
		return (ceil (x - 0.5));
	}
	else {
		return (floor (x + 0.5));
	}
}

#endif

static
uint64_t dbl_to_uint64 (double v)
{
	uint64_t r;

	if (v < 0.0) {
		r = (uint64_t) (-v);
		r = -r;
	}
	else {
		r = (uint64_t) v;
	}

	return (r);
}

static
int dbl_bit_and (double *v, double x, double y)
{
	*v = dbl_to_uint64 (x) & dbl_to_uint64 (y);

	return (0);
}

static
int dbl_bit_or (double *v, double x, double y)
{
	*v = dbl_to_uint64 (x) | dbl_to_uint64 (y);

	return (0);
}

static
int dbl_bit_xor (double *v, double x, double y)
{
	*v = dbl_to_uint64 (x) ^ dbl_to_uint64 (y);

	return (0);
}

static
void set_rnd_seed (unsigned long v)
{
#ifdef HAVE_GETTIMEOFDAY
	struct timeval st;

	if (gettimeofday (&st, NULL) == -1) {
		st.tv_sec = time (NULL);
		st.tv_usec = 0;
	}

	v ^= (unsigned long) st.tv_sec;
	v ^= (unsigned long) st.tv_usec;
#else
	v ^= (unsigned long) time (NULL);
#endif

	mt_init_uint32 (&par_mt, v);

	par_mt_init = 1;
}

static
int set_format (const char *str)
{
	par_format_sign = 0;
	par_format_zero = 0;
	par_format_width = 0;
	par_format_precision = 0;
	par_format_mode = "f";

	if (*str == '+') {
		par_format_sign = 1;
		str += 1;
	}

	if (*str == '0') {
		par_format_zero = 1;
		str += 1;
	}

	while ((*str >= '0') && (*str <= '9')) {
		par_format_width = 10 * par_format_width + (*str - '0');
		str += 1;
	}

	if (*str == '.') {
		str += 1;
	}

	while ((*str >= '0') && (*str <= '9')) {
		par_format_precision = 10 * par_format_precision + (*str - '0');
		str += 1;
	}

	if (*str == 'e') {
		par_format_mode = "e";
		str += 1;
	}
	else if (*str == 'E') {
		par_format_mode = "E";
		str += 1;
	}
	else if (*str == 'f') {
		par_format_mode = "f";
		str += 1;
	}
	else if (*str == 'F') {
		par_format_mode = "F";
		str += 1;
	}
	else if (*str == 'g') {
		par_format_mode = "g";
		str += 1;
	}
	else if (*str == 'G') {
		par_format_mode = "G";
		str += 1;
	}

	if (*str != 0) {
		return (1);
	}

	if ((par_format_width == 0) && (par_format_precision == 0)) {
		par_format_precision = 18;
	}

	return (0);
}

static
int is_digit (char c)
{
	if ((c >= '0') && (c <= '9')) {
		return (1);
	}

	return (0);
}

static
int is_alpha (char c)
{
	if ((c >= 'a') && (c <= 'z')) {
		return (1);
	}

	if ((c >= 'A') && (c <= 'Z')) {
		return (1);
	}

	return (0);
}

static
int match_space (const char **str)
{
	int        r;
	const char *tmp;

	tmp = *str;

	r = 0;
	while ((*tmp == ' ') || (*tmp == '\t')) {
		r = 1;
		tmp += 1;
	}

	*str = tmp;

	return (r);
}

static
int match_string (const char **str, const char *val)
{
	const char *tmp;

	tmp = *str;

	match_space (&tmp);

	while (*val != 0) {
		if (*tmp != *val) {
			return (0);
		}

		tmp += 1;
		val += 1;
	}

	*str = tmp;

	return (1);
}

static
int match_peek_string (const char **str, const char *val)
{
	const char *tmp;

	tmp = *str;

	match_space (&tmp);

	while (*val != 0) {
		if (*tmp != *val) {
			return (0);
		}

		tmp += 1;
		val += 1;
	}

	return (1);
}

static
int match_ident (const char **str, char *val, unsigned max)
{
	unsigned   i;
	const char *tmp;

	tmp = *str;

	match_space (&tmp);

	if ((is_alpha (*tmp) == 0) && (*tmp != '_')) {
		return (0);
	}

	if (max < 2) {
		return (0);
	}

	i = 0;
	val[i++] = *(tmp++);

	while (is_alpha (*tmp) || is_digit (*tmp) || (*tmp == '_')) {
		if (i >= max) {
			return (0);
		}

		val[i++] = *(tmp++);
	}

	if (i >= max) {
		return (0);
	}

	val[i] = 0;

	*str = tmp;

	return (1);
}

static
int match_double (const char **str, double *val)
{
	const char *tmp;
	double     v;

	tmp = *str;

	match_space (&tmp);

	if (is_digit (*tmp) == 0) {
		return (0);
	}

	v = 0.0;
	while (is_digit (*tmp)) {
		v = 10.0 * v + (*tmp - '0');
		tmp += 1;
	}

	if (*tmp == '.') {
		double f, d;

		tmp += 1;

		f = 0.0;
		d = 1.0;

		while (is_digit (*tmp)) {
			f = 10.0 * f + (*tmp - '0');
			d = 10.0 * d;
			tmp += 1;
		}

		v = v + (f / d);
	}

	*val = v;
	*str = tmp;

	return (1);
}

static
int evalf_function_0 (const char *name, double *val)
{
	if (strcmp (name, "pi") == 0) {
		*val = 2.0 * asin (1.0);
	}
	else {
		return (1);
	}

	return (0);
}

static
int evalf_function_1 (const char *name, double *val, double x)
{
	if (strcmp (name, "abs") == 0) {
		*val = fabs (x);
	}
	else if (strcmp (name, "acos") == 0) {
		*val = acos (x);
	}
	else if (strcmp (name, "asin") == 0) {
		*val = asin (x);
	}
	else if (strcmp (name, "ceil") == 0) {
		*val = ceil (x);
	}
	else if (strcmp (name, "cosh") == 0) {
		*val = cosh (x);
	}
	else if (strcmp (name, "cos") == 0) {
		*val = cos (x);
	}
	else if (strcmp (name, "cot") == 0) {
		*val = cos (x) / sin (x);
	}
	else if (strcmp (name, "exp") == 0) {
		*val = exp (x);
	}
	else if (strcmp (name, "floor") == 0) {
		*val = floor (x);
	}
	else if ((strcmp (name, "log") == 0) || (strcmp (name, "ln") == 0)) {
		if (x <= 0.0) {
			return (1);
		}

		*val = log (x);
	}
	else if (strcmp (name, "log2") == 0) {
		if (x <= 0.0) {
			return (1);
		}

		*val = log (x) / log (2.0);
	}
	else if (strcmp (name, "log10") == 0) {
		if (x <= 0.0) {
			return (1);
		}

		*val = log10 (x);
	}
	else if ((strcmp (name, "round") == 0) || (strcmp (name, "int") == 0)) {
		*val = round (x);
	}
	else if (strcmp (name, "sinh") == 0) {
		*val = sinh (x);
	}
	else if (strcmp (name, "sin") == 0) {
		*val = sin (x);
	}
	else if (strcmp (name, "sqrt") == 0) {
		if (x < 0.0) {
			return (1);
		}
		*val = sqrt (x);
	}
	else if (strcmp (name, "sqr") == 0) {
		*val = x * x;
	}
	else if (strcmp (name, "tanh") == 0) {
		*val = tanh (x);
	}
	else if (strcmp (name, "tan") == 0) {
		*val = tan (x);
	}
	else if (strcmp (name, "trunc") == 0) {
		*val = trunc (x);
	}
	else {
		return (1);
	}

	return (0);
}

static
int evalf_function_n (const char *name, double *val, const double *x, unsigned cnt)
{
	unsigned i;

	if (strcmp (name, "atan") == 0) {
		if (cnt == 1) {
			*val = atan (x[0]);
		}
		else if (cnt == 2) {
			*val = atan2 (x[0], x[1]);
		}
		else {
			return (1);
		}
	}
	else if (strcmp (name, "max") == 0) {
		if (cnt == 0) {
			return (1);
		}

		*val = x[0];

		for (i = 1; i < cnt; i++) {
			if (x[i] > *val) {
				*val = x[i];
			}
		}
	}
	else if (strcmp (name, "min") == 0) {
		if (cnt == 0) {
			return (1);
		}

		*val = x[0];

		for (i = 1; i < cnt; i++) {
			if (x[i] < *val) {
				*val = x[i];
			}
		}
	}
	else if ((strcmp (name, "rnd") == 0) || (strcmp (name, "U") == 0)) {
		double x1, x2;

		if (par_mt_init == 0) {
			set_rnd_seed (0);
		}

		if (cnt == 0) {
			x1 = 0.0;
			x2 = 1.0;
		}
		else if (cnt == 1) {
			x1 = 0.0;
			x2 = x[0];
		}
		else if (cnt == 2) {
			x1 = x[0];
			x2 = x[1];
		}
		else {
			return (1);
		}

		*val = x1 + (x2 - x1) * mt_get_double (&par_mt);
	}
	else if (strcmp (name, "N") == 0) {
		static double buf = 0.0;
		static int    valid = 0;
		double        m, s, x1, x2;

		if (par_mt_init == 0) {
			set_rnd_seed (0);
		}

		if (cnt == 0) {
			m = 0.0;
			s = 1.0;
		}
		else if (cnt == 1) {
			m = 0.0;
			s = x[0];
		}
		else if (cnt == 2) {
			m = x[0];
			s = x[1];
		}
		else {
			return (1);
		}

		if (valid) {
			valid = 0;
			*val = m + s * buf;
			return (0);
		}

		x1 = mt_get_double (&par_mt);
		x2 = mt_get_double (&par_mt);

		if (x1 == 0.0) {
			x1 = 1.0;
		}

		x1 = sqrt (-2.0 * log (x1));
		x2 = 2.0 * 3.141592653589793116 * x2;

		*val = m + s * x1 * cos (x2);

		buf = x1 * sin (x2);
		valid = 1;
	}
	else {
		return (1);
	}

	return (0);
}

static
int evalf_ident (const char **str, double *val, const char *name)
{
	unsigned cnt;
	double   x[64];

	cnt = 0;

	if (match_string (str, "(")) {
		while (cnt < 64) {
			if (evalf_expr (str, &x[cnt])) {
				return (1);
			}

			cnt += 1;

			if (match_string (str, ")")) {
				break;
			}

			if (match_string (str, ",") == 0) {
				return (1);
			}
		}
	}

	if (evalf_function_n (name, val, x, cnt) == 0) {
		return (0);
	}

	if (cnt == 0) {
		if (evalf_function_0 (name, val) == 0) {
			return (0);
		}
	}
	else if (cnt == 1) {
		if (evalf_function_1 (name, val, x[0]) == 0) {
			return (0);
		}
	}

	return (1);
}

static
int evalf_literal (const char **str, double *val)
{
	char name[256];

	if (match_double (str, val)) {
		return (0);
	}

	if (match_string (str, "(")) {
		if (evalf_expr (str, val)) {
			return (1);
		}

		if (match_string (str, ")") == 0) {
			return (1);
		}

		return (0);
	}

	if (match_ident (str, name, 256)) {
		return (evalf_ident (str, val, name));
	}

	return (1);
}

static
int evalf_exp (const char **str, double *val)
{
	double v;

	if (evalf_literal (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "**")) {
			if (evalf_literal (str, &v)) {
				return (1);
			}
			*val = pow (*val, v);
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_neg (const char **str, double *val)
{
	if (match_string (str, "+")) {
		if (evalf_neg (str, val)) {
			return (1);
		}
		return (0);
	}
	else if (match_string (str, "-")) {
		if (evalf_neg (str, val)) {
			return (1);
		}
		*val = -*val;
		return (0);
	}
	else if (match_string (str, "!") || match_string (str, "~")) {
		if (evalf_neg (str, val)) {
			return (1);
		}
		*val = (*val < 0.0) ? 1.0 : -1.0;
		return (0);
	}

	if (evalf_exp (str, val)) {
		return (1);
	}

	return (0);
}

static
int evalf_mul (const char **str, double *val)
{
	double v;

	if (evalf_neg (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "*")) {
			if (evalf_neg (str, &v)) {
				return (1);
			}
			*val = *val * v;
		}
		else if (match_string (str, "/")) {
			if (evalf_neg (str, &v)) {
				return (1);
			}
			if (v == 0.0) {
				return (1);
			}
			*val = *val / v;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_add (const char **str, double *val)
{
	double v;

	if (evalf_mul (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "+")) {
			if (evalf_mul (str, &v)) {
				return (1);
			}
			*val = *val + v;
		}
		else if (match_string (str, "-")) {
			if (evalf_mul (str, &v)) {
				return (1);
			}
			*val = *val - v;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_shift (const char **str, double *val)
{
	double v;

	if (evalf_add (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "<<")) {
			if (evalf_add (str, &v)) {
				return (1);
			}
			*val = *val * pow (2.0, v);
		}
		else if (match_string (str, ">>")) {
			if (evalf_add (str, &v)) {
				return (1);
			}
			*val = *val * pow (2.0, -v);
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_cmp (const char **str, double *val)
{
	double v;

	if (evalf_shift (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "<=")) {
			if (evalf_shift (str, &v)) {
				return (1);
			}
			*val = (*val <= v) ? 1.0 : -1.0;
		}
		else if (match_string (str, ">=")) {
			if (evalf_shift (str, &v)) {
				return (1);
			}
			*val = (*val >= v) ? 1.0 : -1.0;
		}
		else if (match_peek_string (str, "<<")) {
			break;
		}
		else if (match_peek_string (str, ">>")) {
			break;
		}
		else if (match_string (str, "<")) {
			if (evalf_shift (str, &v)) {
				return (1);
			}
			*val = (*val < v) ? 1.0 : -1.0;
		}
		else if (match_string (str, ">")) {
			if (evalf_shift (str, &v)) {
				return (1);
			}
			*val = (*val > v) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_equ (const char **str, double *val)
{
	double v;

	if (evalf_cmp (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "==")) {
			if (evalf_cmp (str, &v)) {
				return (1);
			}
			*val = (*val == v) ? 1.0 : -1.0;
		}
		else if (match_string (str, "!=")) {
			if (evalf_cmp (str, &v)) {
				return (1);
			}
			*val = (*val != v) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_bit_and (const char **str, double *val)
{
	double v;

	if (evalf_equ (str, val)) {
		return (1);
	}

	while (1) {
		if (match_peek_string (str, "&&")) {
			break;
		}

		if (match_string (str, "&")) {
			if (evalf_equ (str, &v)) {
				return (1);
			}
			if (dbl_bit_and (val, *val, v)) {
				return (1);
			}
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_bit_xor (const char **str, double *val)
{
	double v;

	if (evalf_bit_and (str, val)) {
		return (1);
	}

	while (1) {
		if (match_peek_string (str, "^^")) {
			break;
		}

		if (match_string (str, "^")) {
			if (evalf_bit_and (str, &v)) {
				return (1);
			}
			if (dbl_bit_xor (val, *val, v)) {
				return (1);
			}
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_bit_or (const char **str, double *val)
{
	double v;

	if (evalf_bit_xor (str, val)) {
		return (1);
	}

	while (1) {
		if (match_peek_string (str, "||")) {
			break;
		}

		if (match_string (str, "|")) {
			if (evalf_bit_xor (str, &v)) {
				return (1);
			}
			if (dbl_bit_or (val, *val, v)) {
				return (1);
			}
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_and (const char **str, double *val)
{
	double v;

	if (evalf_bit_or (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "&&")) {
			if (evalf_bit_or (str, &v)) {
				return (1);
			}
			*val = ((*val > 0.0) && (v > 0.0)) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_xor (const char **str, double *val)
{
	double v;

	if (evalf_and (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "^^")) {
			if (evalf_and (str, &v)) {
				return (1);
			}
			*val = ((*val > 0.0) != (v > 0.0)) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_or (const char **str, double *val)
{
	double v;

	if (evalf_xor (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "||")) {
			if (evalf_xor (str, &v)) {
				return (1);
			}
			*val = ((*val > 0.0) || (v > 0.0)) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_if (const char **str, double *val)
{
	double v1, v2;

	if (evalf_or (str, val)) {
		return (1);
	}

	if (match_string (str, "?")) {
		if (evalf_or (str, &v1)) {
			return (1);
		}

		if (match_string (str, ":") == 0) {
			return (1);
		}

		if (evalf_or (str, &v2)) {
			return (1);
		}

		*val = (*val >= 0.0) ? v1 : v2;
	}

	return (0);
}

static
int evalf_expr (const char **str, double *val)
{
	if (evalf_if (str, val)) {
		return (1);
	}

	return (0);
}

static
int evalf_str (const char *str)
{
	double     val;
	const char *fmt;
	char       buf[64];

	val = 0.0;

	if (evalf_expr (&str, &val)) {
		return (1);
	}

	if ((par_format_width == 0) && (par_format_precision == 0)) {
		fmt = "%.18g\n";
	}
	else {
		sprintf (buf, "%%%s%s%u.%u%s\n",
			par_format_sign ? "+" : "",
			par_format_zero ? "0" : "",
			par_format_width,
			par_format_precision,
			par_format_mode
		);
		fmt = buf;
	}

	fprintf (stdout, fmt, val);

	return (0);
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;

	argv0 = argv[0];

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("evalf");
			return (EXIT_SUCCESS);

		case 'e':
			if (evalf_str (optarg[0])) {
				fprintf (stderr, "%s: bad expression (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'f':
			if (set_format (optarg[0])) {
				fprintf (stderr, "%s: bad format (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 0:
			if (evalf_str (optarg[0])) {
				fprintf (stderr, "%s: bad expression (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	return (EXIT_SUCCESS);
}
