/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/fcmp/fcmp.c                                              *
 * Created:     2000-05-12 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#if defined(HAVE_TIME_H) && defined(HAVE_UNISTD_H)
#if defined(HAVE_SYS_TYPES_H) && defined(HAVE_SYS_STAT_H)
#define FCMP_STAT 1
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif
#endif

#include "fcmp.h"


typedef struct {
	FILE          *fp;
	unsigned char chr;
	int           eof;
	char          *fname;
} fcmp_t;


static const char    *argv0 = NULL;

static int           par_quiet = 0;
static int           par_header = 0;
static unsigned long par_diff_max = 0;

static unsigned long par_file_cnt = 0;
static fcmp_t        *par_files = NULL;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'f', 0, "first", NULL, "Compare only up to the first difference [no]" },
	{ 'h', 0, "header", NULL, "Print a header [no]" },
	{ 'i', 1, "input", "string", "Set an input file name [none]" },
	{ 'n', 1, "max-diff", "n", "Stop after n different bytes [0]" },
	{ 'q', 0, "quiet", NULL, "No output to stdout [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"fcmp: compare binary files",
		"usage: fcmp [options] [filename...]",
		opts
	);

	fflush (stdout);
}

static
int fcmp_init_file (fcmp_t *fc, const char *fname)
{
	unsigned n;

	fc->fp = NULL;
	fc->chr = 0;
	fc->eof = 0;
	fc->fname = NULL;

	if (strcmp (fname, "-") == 0) {
		fc->fp = misc_open_inp (argv0, fname, "rb");
		return (0);
	}

	n = strlen (fname);

	fc->fname = malloc (n + 1);

	if (fc->fname == NULL) {
		return (1);
	}

	memcpy (par_files[par_file_cnt].fname, fname, n + 1);

	fc->fp = fopen (fname, "rb");

	if (fc->fp == NULL) {
		free (fc->fname);
		return (1);
	}

	return (0);
}

static
int fcmp_new_file (const char *fname)
{
	fcmp_t *fc;

	fc = realloc (par_files, sizeof (fcmp_t) * (par_file_cnt + 1));

	if (fc == NULL) {
		return (1);
	}

	par_files = fc;

	if (fcmp_init_file (fc + par_file_cnt, fname)) {
		return (1);
	}

	par_file_cnt += 1;

	return (0);
}

static
void fcmp_close (void)
{
	unsigned long i;

	for (i = 0; i < par_file_cnt; i++) {
		misc_close (par_files[i].fp);
		free (par_files[i].fname);
	}
}

/* read next characters. returns 1 if all files have reached eof. */
static
int fcmp_next_char (void)
{
	unsigned long i;
	int           c;
	int           eof;

	eof = 1;

	for (i = 0; i < par_file_cnt; i++) {
		if (par_files[i].eof) {
			continue;
		}

		c = fgetc (par_files[i].fp);

		if (c == EOF) {
			par_files[i].eof = 1;
		}
		else {
			par_files[i].chr = c;
			eof = 0;
		}
	}

	return (eof);
}

/* compare current characters. returns 0 if all characters are equal */
static
int fcmp_test_char (void)
{
	unsigned long i;

	for (i = 1; i < par_file_cnt; i++) {
		if (par_files[0].eof != par_files[i].eof) {
			return (1);
		}

		if (par_files[i].eof == 0) {
			if (par_files[0].chr != par_files[i].chr) {
				return (1);
			}
		}
	}

	return (0);
}

static
void fcmp_prt_char (FILE *fp, unsigned long ofs)
{
	unsigned long i;

	fprintf (fp, "%08lX", ofs);

	for (i = 0; i < par_file_cnt; i++) {
		if (par_files[i].eof) {
			fputs (" EOF", fp);
		}
		else {
			fprintf (fp, "  %02X", (unsigned) par_files[i].chr);
		}
	}

	fputs ("\n", fp);

	fflush (fp);
}

static
void fcmp_prt_header (FILE *fp)
{
	unsigned i;

	for (i = 0; i < par_file_cnt; i++) {
		if (par_files[i].fname == NULL) {
			fprintf (fp, "# %u: <stdin>\n", i + 1);
		}
		else {
#ifdef FCMP_STAT
			struct stat st;
			struct tm   *tm;

			if (stat (par_files[i].fname, &st)) {
				fprintf (fp, "# %u: %s\n", i + 1, par_files[i].fname);
			}
			else {
				tm = localtime (&st.st_mtime);

				fprintf (fp, "# %u: %s %04d-%02d-%02d %02d:%02d:%02d %lu\n", i + 1,
					par_files[i].fname,
					1900 + tm->tm_year, tm->tm_mon + 1, tm->tm_mday,
					tm->tm_hour, tm->tm_min, tm->tm_sec,
					(unsigned long) st.st_size
				);
			}
#else
		fprintf (fp, "# %u: %s\n", i + 1, par_files[i].fname);
#endif
		}
	}

	fflush (fp);
}

/* compare files. returns 0 if all files equal */
static
int fcmp_compare (void)
{
	int           equ;
	unsigned long cnt;
	unsigned long ofs;

	if (par_header) {
		fcmp_prt_header (stdout);
	}

	equ = 1;
	cnt = 0;
	ofs = 0;

	while (1) {
		if (fcmp_next_char()) {
			return (!equ);
		}

		if (fcmp_test_char()) {
			equ = 0;

			if (!par_quiet) {
				fcmp_prt_char (stdout, ofs);
			}

			cnt += 1;

			if ((par_diff_max > 0) && (cnt >= par_diff_max)) {
				return (!equ);
			}
		}

		ofs += 1;
	}

	return (!equ);
}

int main (int argc, char **argv)
{
	int  r;
	char **optarg;

	argv0 = argv[0];

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("fcmp");
			return (EXIT_SUCCESS);

		case 'f':
			par_diff_max = 1;
			break;

		case 'h':
			par_header = 1;
			break;

		case 'i':
			if (fcmp_new_file (optarg[0])) {
				fprintf (stderr,
					"%s: can't open input file (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'n':
			if (str_get_ulng (optarg[0], &par_diff_max)) {
				fprintf (stderr,
					"%s: bad max diff count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'q':
			par_quiet = 1;
			break;

		case 0:
			if (fcmp_new_file (optarg[0])) {
				fprintf (stderr,
					"%s: can't open input file (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	r = fcmp_compare();

	fcmp_close();

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
