/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/hex/hex.c                                                *
 * Created:     2000-12-07 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include "hex.h"


static const char    *argv0 = NULL;

static unsigned      par_dump = 1;
static int           par_dump_c = 0;
static unsigned      par_cols = 16;
static unsigned      par_base = 16;
static unsigned      par_digits = 2;
static int           par_chars = 1;
static int           par_lcase = 0;
static unsigned      par_spaces = 1;
static unsigned long par_addr = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 1, "address", "int",  "Set the base address [0]" },
	{ 'b', 1, "base", "int", "Set base for dump [16]" },
	{ 'c', 1, "columns", "int", "Set number of bytes per row [16]" },
	{ 'C', 0, "dump-c", NULL, "Use a C style dump format [no]" },
	{ 'd', 0, "decode", NULL, "Reconstruct a binary file from a hex dump" },
	{ 'e', 0, "encode", NULL, "Dump a binary file [default]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'l', 0, "lower-case", NULL, "Use lower case letters [no]" },
	{ 'n', 1, "digits", "int", "Set number of digits per byte [2]" },
	{ 'o', 1, "output", "string", "Set output file name [stdout]" },
	{ 'p', 1, "spaces", "int", "Set number of spaces between bytes [1]" },
	{ 's', 0, "short", NULL, "Select short output format [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"hex: hex dump of binary files",
		"usage: hex [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
int is_digit (unsigned c)
{
	unsigned d;

	d = par_base;

	if ((c >= '0') && (c <= '9')) {
		d = (unsigned)(c - '0');
	}
	else if ((c >= 'a') && (c <= 'z')) {
		d = (unsigned)(10 + c - 'a');
	}
	else if ((c >= 'A') && (c <= 'Z')) {
		d = (unsigned)(10 + c - 'A');
	}

	return (d < par_base);
}

static
unsigned dig_to_int (unsigned char c)
{
	if ((c >= '0') && (c <= '9')) {
		return (c - '0');
	}
	else if ((c >= 'a') && (c <= 'z')) {
		return (c + 10 - 'a');
	}
	else if ((c >= 'A') && (c <= 'Z')) {
		return (c + 10 - 'A');
	}

	return (0);
}

static
void set_char (FILE *fp, unsigned c)
{
	unsigned i, d;
	char     buf[HEX_MAX_DIG];

	i = 0;

	do {
		d = c % par_base;
		c = c / par_base;

		if (par_lcase) {
			d += (d < 10) ? '0' : ('a' - 10);
		}
		else {
			d += (d < 10) ? '0' : ('A' - 10);
		}

		buf[i++] = (char) d;
		if (i >= HEX_MAX_DIG) {
			break;
		}
	} while ((c != 0) || (i < par_digits));

	while (i > 0) {
		fputc (buf[--i], fp);
	}
}

static
void set_char_c (FILE *fp, unsigned c)
{
	fprintf (fp, "0x%02x", c & 0xff);
}

static
void set_line (FILE *fp, unsigned long ofs, unsigned char *ln, unsigned cnt)
{
	unsigned i, j;

	if (cnt == 0) {
		return;
	}

	for (i = 1; i < cnt; i++) {
		set_char (fp, ln[i - 1]);
		for (j = 0; j < par_spaces; j++) {
			fputc (' ', fp);
		}
	}

	set_char (fp, ln[cnt - 1]);

	if (par_chars) {
		fprintf (fp, "\t# %08lX  ", par_addr + ofs);

		for (i = 0; i < cnt; i++) {
			if ((ln[i] >= 32) && (ln[i] < 127)) {
				fputc (ln[i], fp);
			}
			else {
				fputc ('.', fp);
			}
		}
	}

	fputc ('\n', fp);
}

static
void set_line_c (FILE *fp, unsigned long ofs, unsigned char *ln, unsigned cnt,
	int last)
{
	unsigned i;

	if (cnt == 0) {
		return;
	}

	for (i = 1; i < cnt; i++) {
		set_char_c (fp, ln[i - 1]);
		fputs (", ", fp);
	}

	set_char_c (fp, ln[cnt - 1]);

	if (last == 0) {
		fputs (",", fp);
	}

	if (par_chars) {
		fprintf (fp, "\t/* %08lX */", ofs);
	}

	fputc ('\n', fp);
}

static
int do_dump (FILE *inp, FILE *out)
{
	unsigned      col;
	unsigned long ofs;
	unsigned char buf[HEX_MAX_COL];
	int           next;

	ofs = 0;

	next = fgetc (inp);

	while (next != EOF) {
		buf[0] = (unsigned char) next;
		col = 1 + fread (buf + 1, 1, par_cols - 1, inp);
		next = fgetc (inp);

		if (par_dump_c) {
			set_line_c (out, ofs, buf, col, next == EOF);
		}
		else {
			set_line (out, ofs, buf, col);
		}

		ofs += col;
	}

	fflush (out);

	return (0);
}

static
int do_dump_stream (FILE *inp, FILE *out)
{
	int      c;
	unsigned i;

	c = fgetc (inp);
	if (c != EOF) {
		set_char (out, c);
	}

	c = fgetc (inp);
	while (c != EOF) {
		for (i = 0; i < par_spaces; i++) {
			fputc (' ', out);
		}

		set_char (out, c);

		c = fgetc (inp);
	}

	fflush (out);

	return (0);
}

static
int undump_digit (FILE *inp, FILE *out, int c)
{
	unsigned i, v;

	i = 0;
	v = 0;

	while (is_digit (c)) {
		v = (par_base * v + dig_to_int (c)) & 0xff;

		i += 1;
		if ((par_digits > 0) && (i >= par_digits)) {
			break;
		}

		c = fgetc (inp);
		if (c == EOF) {
			break;
		}
	}

	fputc (v, out);

	return (0);
}

static
int undump_string (FILE *inp, FILE *out)
{
	int c;

	while (1) {
		c = fgetc (inp);
		if (c == EOF) {
			return (1);
		}

		if (c == '"') {
			return (0);
		}

		fputc (c, out);
	}

	return (0);
}

static
int do_undump (FILE *inp, FILE *out)
{
	int c;

	while (1) {
		c = fgetc (inp);
		if (c == EOF) {
			break;
		}

		if (is_digit (c)) {
			if (undump_digit (inp, out, c)) {
				return (1);
			}
		}
		else if (c == '"') {
			if (undump_string (inp, out)) {
				return (1);
			}
		}
		else if (c == '#') {
			do {
				c = fgetc (inp);
			} while ((c != 0x0a) && (c != 0x0d) && (c != EOF));
		}
	}

	fflush (out);

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	int        have_digits;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	have_digits = 0;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("hex");
			return (EXIT_SUCCESS);

		case 'a':
			if (str_get_ulng (optarg[0], &par_addr)) {
				fprintf (stderr, "%s: bad address (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'b':
			if (str_get_uint (optarg[0], &par_base)) {
				fprintf (stderr, "%s: bad base (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'c':
			if (str_get_uint (optarg[0], &par_cols)) {
				fprintf (stderr, "%s: bad column count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'C':
			par_dump_c = 1;
			break;

		case 'd':
			par_dump = 0;
			if (have_digits == 0) {
				par_digits = 0;
			}
			break;

		case 'e':
			par_dump = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'l':
			par_lcase = 1;
			break;

		case 'n':
			if (str_get_uint (optarg[0], &par_digits)) {
				fprintf (stderr, "%s: bad digit count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			have_digits = 1;
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			if (str_get_uint (optarg[0], &par_spaces)) {
				fprintf (stderr, "%s: bad spaces count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 's':
			par_chars = 0;
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_cols > HEX_MAX_COL) {
		par_cols = HEX_MAX_COL;
	}

	if (par_base < 2) {
		par_base = 2;
	}
	else if (par_base > 36) {
		par_base = 36;
	}

	if (par_digits > HEX_MAX_DIG) {
		par_digits = HEX_MAX_DIG;
	}

	inp = misc_open_inp (argv0, finp, par_dump ? "rb" : "r");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, par_dump ? "w" : "wb");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	if (par_dump) {
		if (par_cols == 0) {
			r = do_dump_stream (inp, out);
		}
		else {
			r = do_dump (inp, out);
		}
	}
	else {
		r = do_undump (inp, out);
	}

	misc_close (stdout);
	misc_close (stdin);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
