/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/md4.c                                                *
 * Created:     2004-07-08 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/md4.h>


#define md4_f(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define md4_g(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define md4_h(x, y, z) ((x) ^ (y) ^ (z))

#define md4_rol(x, n) (((x) << (n)) | ((x) >> (32 - (n))))

#define md4_r1(a, b, c, d, M, s) do { \
	(a) = ((a) + md4_f ((b), (c), (d)) + (M)) & 0xffffffff; \
	(a) = md4_rol ((a), (s)) & 0xffffffff; \
	} while (0)

#define md4_r2(a, b, c, d, M, s) do { \
	(a) = ((a) + md4_g ((b), (c), (d)) + (M) + 0x5a827999) & 0xffffffff; \
	(a) = md4_rol ((a), (s)) & 0xffffffff; \
	} while (0)

#define md4_r3(a, b, c, d, M, s) do { \
	(a) = ((a) + md4_h ((b), (c), (d)) + (M) + 0x6ed9eba1) & 0xffffffff; \
	(a) = md4_rol ((a), (s)) & 0xffffffff; \
	} while (0)


void md4_init (md4_t *md4)
{
	md4->buf_i = 0;

	md4->cnt[0] = 0;
	md4->cnt[1] = 0;

	md4->h[0] = 0x67452301UL;
	md4->h[1] = 0xefcdab89UL;
	md4->h[2] = 0x98badcfeUL;
	md4->h[3] = 0x10325476UL;
}

void md4_copy (md4_t *dst, const md4_t *src)
{
	unsigned i;

	dst->buf_i = src->buf_i;

	for (i = 0; i < src->buf_i; i++) {
		dst->buf[i] = src->buf[i];
	}

	dst->cnt[0] = src->cnt[0];
	dst->cnt[1] = src->cnt[1];

	for (i = 0; i < 4; i++) {
		dst->h[i] = src->h[i];
	}
}

static
void md4_set_uint32 (void *buf, unsigned i, unsigned long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = v & 0xff;
	tmp[1] = (v >> 8) & 0xff;
	tmp[2] = (v >> 16) & 0xff;
	tmp[3] = (v >> 24) & 0xff;
}

static
void md4_add_size (md4_t *md4, unsigned long n)
{
	n &= 0xffffffffUL;

	md4->cnt[0] = (md4->cnt[0] + n) & 0xffffffffUL;

	if (md4->cnt[0] < n) {
		md4->cnt[1] = (md4->cnt[1] + 1) & 0xffffffffUL;
	}
}

static
void md4_blk_frm_buf (unsigned long *dst, const unsigned char *src)
{
	unsigned      i;
	unsigned long v;

	for (i = 0; i < 16; i++) {
		v = src[3];
		v = (v << 8) + src[2];
		v = (v << 8) + src[1];
		v = (v << 8) + src[0];

		dst[i] = v;

		src += 4;
	}
}

static
void md4_calc_block (md4_t *md4, const void *buf)
{
	unsigned long a, b, c, d;
	unsigned long M[16];

	md4_blk_frm_buf (M, buf);

	a = md4->h[0];
	b = md4->h[1];
	c = md4->h[2];
	d = md4->h[3];

	md4_r1 (a, b, c, d, M[ 0],  3);
	md4_r1 (d, a, b, c, M[ 1],  7);
	md4_r1 (c, d, a, b, M[ 2], 11);
	md4_r1 (b, c, d, a, M[ 3], 19);
	md4_r1 (a, b, c, d, M[ 4],  3);
	md4_r1 (d, a, b, c, M[ 5],  7);
	md4_r1 (c, d, a, b, M[ 6], 11);
	md4_r1 (b, c, d, a, M[ 7], 19);
	md4_r1 (a, b, c, d, M[ 8],  3);
	md4_r1 (d, a, b, c, M[ 9],  7);
	md4_r1 (c, d, a, b, M[10], 11);
	md4_r1 (b, c, d, a, M[11], 19);
	md4_r1 (a, b, c, d, M[12],  3);
	md4_r1 (d, a, b, c, M[13],  7);
	md4_r1 (c, d, a, b, M[14], 11);
	md4_r1 (b, c, d, a, M[15], 19);

	md4_r2 (a, b, c, d, M[ 0],  3);
	md4_r2 (d, a, b, c, M[ 4],  5);
	md4_r2 (c, d, a, b, M[ 8],  9);
	md4_r2 (b, c, d, a, M[12], 13);
	md4_r2 (a, b, c, d, M[ 1],  3);
	md4_r2 (d, a, b, c, M[ 5],  5);
	md4_r2 (c, d, a, b, M[ 9],  9);
	md4_r2 (b, c, d, a, M[13], 13);
	md4_r2 (a, b, c, d, M[ 2],  3);
	md4_r2 (d, a, b, c, M[ 6],  5);
	md4_r2 (c, d, a, b, M[10],  9);
	md4_r2 (b, c, d, a, M[14], 13);
	md4_r2 (a, b, c, d, M[ 3],  3);
	md4_r2 (d, a, b, c, M[ 7],  5);
	md4_r2 (c, d, a, b, M[11],  9);
	md4_r2 (b, c, d, a, M[15], 13);

	md4_r3 (a, b, c, d, M[ 0],  3);
	md4_r3 (d, a, b, c, M[ 8],  9);
	md4_r3 (c, d, a, b, M[ 4], 11);
	md4_r3 (b, c, d, a, M[12], 15);
	md4_r3 (a, b, c, d, M[ 2],  3);
	md4_r3 (d, a, b, c, M[10],  9);
	md4_r3 (c, d, a, b, M[ 6], 11);
	md4_r3 (b, c, d, a, M[14], 15);
	md4_r3 (a, b, c, d, M[ 1],  3);
	md4_r3 (d, a, b, c, M[ 9],  9);
	md4_r3 (c, d, a, b, M[ 5], 11);
	md4_r3 (b, c, d, a, M[13], 15);
	md4_r3 (a, b, c, d, M[ 3],  3);
	md4_r3 (d, a, b, c, M[11],  9);
	md4_r3 (c, d, a, b, M[ 7], 11);
	md4_r3 (b, c, d, a, M[15], 15);

	md4->h[0] = (md4->h[0] + a) & 0xffffffff;
	md4->h[1] = (md4->h[1] + b) & 0xffffffff;
	md4->h[2] = (md4->h[2] + c) & 0xffffffff;
	md4->h[3] = (md4->h[3] + d) & 0xffffffff;
}

void md4_calc (md4_t *md4, const void *buf, unsigned long cnt)
{
	unsigned            i, n;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	if (md4->buf_i > 0) {
		while ((md4->buf_i < 64) && (cnt > 0)) {
			md4->buf[md4->buf_i++] = *(tmp++);
			cnt -= 1;
		}

		if (md4->buf_i < 64) {
			return;
		}

		md4_calc_block (md4, md4->buf);

		md4->buf_i = 0;
		md4_add_size (md4, 8 * 64);
	}

	while (cnt >= 64) {
		md4_calc_block (md4, tmp);

		md4_add_size (md4, 8 * 64);

		tmp += 64;
		cnt -= 64;
	}

	n = (unsigned) cnt;
	for (i = 0; i < n; i++) {
		md4->buf[i] = tmp[i];
	}

	md4->buf_i = n;
}

void md4_done (md4_t *md4)
{
	unsigned i;

	md4_add_size (md4, 8 * md4->buf_i);

	md4->buf[md4->buf_i++] = 0x80;

	if (md4->buf_i > 56) {
		for (i = md4->buf_i; i < 64; i++) {
			md4->buf[i] = 0;
		}

		md4_calc_block (md4, md4->buf);

		md4->buf_i = 0;
	}

	for (i = md4->buf_i; i < 56; i++) {
		md4->buf[i] = 0;
	}

	md4_set_uint32 (md4->buf, 56, md4->cnt[0]);
	md4_set_uint32 (md4->buf, 60, md4->cnt[1]);

	md4_calc_block (md4, md4->buf);
}

unsigned long md4_get_sum_i (md4_t *md4, unsigned i)
{
	if (i < 4) {
		return (md4->h[i]);
	}

	return (0);
}

void md4_get_sum (md4_t *md4, void *buf)
{
	md4_set_uint32 (buf, 0, md4->h[0]);
	md4_set_uint32 (buf, 4, md4->h[1]);
	md4_set_uint32 (buf, 8, md4->h[2]);
	md4_set_uint32 (buf, 12, md4->h[3]);
}

void md4_calc1 (void *dst, const void *src, unsigned long n)
{
	md4_t md4;

	md4_init (&md4);
	md4_calc (&md4, src, n);
	md4_done (&md4);
	md4_get_sum (&md4, dst);
}
