/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/rate/rate.c                                              *
 * Created:     2003-12-07 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2003-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/strtoint.h>

#include <unistd.h>
#include <sys/time.h>

#include "rate.h"


#define RATE_BUF_SIZE MISC_BUF_SIZE


static const char         *argv0 = NULL;

static int                par_limit = 0;
static int                par_verbose = 0;

static unsigned long      par_delay = 1000;

static unsigned long long par_t0 = 0;
static unsigned long long par_cnt = 0;

static unsigned long      par_num = 0;
static unsigned long      par_den = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'd', 1, "delay", "int", "Set the update delay in milliseconds [1000]" },
	{ 'r', 2, "rate", "cnt ms", "Limit the rate to cnt bytes per ms milliseconds" },
	{ 'v', 0, "verbose", NULL, "Print the current rate [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"rate: limit the data rate in a pipe",
		"usage: rate [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
unsigned long long time_get (void)
{
	unsigned long long val;
	struct timeval     tv;

	if (gettimeofday (&tv, NULL)) {
		return (0);
	}

	val = (tv.tv_sec * 1000ULL) + (tv.tv_usec / 1000) - par_t0;

	return (val);
}

static
void time_init (void)
{
	par_t0 = 0;
	par_t0 = time_get();
	par_cnt = 0;
}

static
void time_wait (unsigned long long v)
{
	unsigned long long t;

	while (1) {
		t = time_get();
		if (t >= v) {
			break;
		}

		t = v - t;

		if (t >= 1000) {
			sleep (t / 1000);
		}
		else {
			usleep (1000 * t);
		}
	}
}

static
void prt_size (FILE *fp, unsigned long long n)
{
	if (n < 1024) {
		fprintf (fp, "%.4fB ", (double) n);
	}
	else if (n < (1024ULL * 1024ULL)) {
		fprintf (fp, "%.4fKB", (double) n / 1024.0);
	}
	else if (n < (1024ULL * 1024ULL * 1024ULL)) {
		fprintf (fp, "%.4fMB", (double) n / (1024.0 * 1024.0));
	}
	else {
		fprintf (fp, "%.4fGB", (double) n / (1024.0 * 1024.0 * 1024.0));
	}
}

static
void prt_stat (const char *eol)
{
	unsigned           hrs, min, sec;
	unsigned long long rate;
	unsigned long long dt;

	if (par_verbose == 0) {
		return;
	}

	dt = time_get();

	if (dt > 9) {
		rate = (100 * par_cnt) / (dt / 10);
	}
	else {
		rate = 0;
	}

	sec = (dt / 1000UL) % 60UL;
	min = (dt / (60UL * 1000UL)) % 60UL;
	hrs = (dt / (60UL * 60UL * 1000UL));

	prt_size (stderr, rate);
	fputs ("/s [", stderr);
	prt_size (stderr, par_cnt);
	fprintf (stderr, " / %02u:%02u:%02u]    %s", hrs, min, sec, eol);
	fflush (stderr);
}

static
int do_rate (file_t *inp, file_t *out)
{
	unsigned char      buf[RATE_BUF_SIZE];
	unsigned long      cnt;
	unsigned long long t1, t2;

	time_init();

	cnt = file_read (inp, buf, RATE_BUF_SIZE);

	prt_stat ("\r");

	t1 = time_get();

	while (cnt > 0) {
		if (file_write_block (out, buf, cnt) != cnt) {
			prt_stat ("\n");
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		par_cnt += cnt;

		t2 = time_get();

		if ((t2 - t1) >= par_delay) {
			prt_stat ("\r");
			t1 = t2;
		}

		cnt = file_read (inp, buf, RATE_BUF_SIZE);
	}

	prt_stat ("\n");

	return (0);
}

static
int rate_copy (file_t *inp, file_t *out, unsigned long cnt)
{
	unsigned long n;
	unsigned char buf[RATE_BUF_SIZE];

	while (cnt > 0) {
		n = (cnt < RATE_BUF_SIZE) ? cnt : RATE_BUF_SIZE;

		n = file_read (inp, buf, n);

		if (n == 0) {
			prt_stat ("\n");
			fprintf (stderr, "%s: read error\n", argv0);
			return (1);
		}

		if (file_write_block (out, buf, n) != n) {
			prt_stat ("\n");
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		cnt -= n;
	}

	return (0);
}

static
int do_limit_rate (file_t *inp, file_t *out)
{
	unsigned long long t;
	unsigned long long n1, n2;

	time_init();

	prt_stat ("\r");

	t = time_get();
	n1 = t + par_delay;
	n2 = t + par_den;

	while (1) {
		if (rate_copy (inp, out, par_num)) {
			return (1);
		}

		par_cnt += par_num;

		t = time_get();

		if (t < n2) {
			time_wait (n2);
			n2 += par_den;
		}
		else {
			n2 = t + par_den;
		}

		t = time_get();

		if (t >= n1) {
			prt_stat ("\r");
			n1 = t + par_delay;
		}
	}

	prt_stat ("\n");

	return (0);
}

int main (int argc, char *argv[])
{
	int    r;
	char   **optarg;
	int    have_inp, have_out;
	file_t finp, fout;

	argv0 = argv[0];

	file_set_stdin (&finp);
	file_set_stdout (&fout);

	have_inp = 0;
	have_out = 0;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("rate");
			return (EXIT_SUCCESS);

		case 'd':
			if (str_get_ulng (optarg[0], &par_delay)) {
				fprintf (stderr, "%s: bad update delay (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'i':
			file_set_name (&finp, optarg[0]);
			have_inp = 1;
			break;

		case 'o':
			file_set_name (&fout, optarg[0]);
			have_out = 1;
			break;

		case 'r':
			if (str_get_ulng (optarg[0], &par_num)) {
				fprintf (stderr, "%s: bad byte count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (optarg[1], &par_den)) {
				fprintf (stderr, "%s: bad time (%s)\n",
					argv0, optarg[1]
				);
				return (EXIT_FAILURE);
			}

			par_limit = 1;
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (have_inp == 0) {
				file_set_name (&finp, optarg[0]);
				have_inp = 1;
			}
			else if (have_out == 0) {
				file_set_name (&fout, optarg[0]);
				have_out = 1;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (file_open_inp (&finp)) {
		fprintf (stderr, "%s: can't open input file\n", argv0);
		return (1);
	}

	if (file_open_out (&fout)) {
		fprintf (stderr, "%s: can't open output file\n", argv0);
		return (1);
	}

	if (par_limit) {
		r = do_limit_rate (&finp, &fout);
	}
	else {
		r = do_rate (&finp, &fout);
	}

	file_close (&fout);
	file_close (&finp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
