/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/readbin/readbin.c                                        *
 * Created:     2002-05-05 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include <stdlib.h>

#include "readbin.h"


#define MISC_READBIN_BUFFER MISC_BUF_SIZE


static const char *argv0 = NULL;

static unsigned par_base = 10;
static int      par_little = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'b', 0, "big-endian", NULL, "Use big endian [default]" },
	{ 'c', 0, "cat", NULL, "Read rest of input file" },
	{ 'i', 1, "input", "string", "Set an input file name [stdin]" },
	{ 'k', 1, "skip", "n", "Skip n bytes" },
	{ 'l', 0, "little-endian", NULL, "Use little endian" },
	{ 'n', 1, "base", "int", "Set number system base [10]" },
	{ 'o', 1, "output", "string", "Set an output file name [stdout]" },
	{ 'p', 1, "position", "int", "Set absolute file position" },
	{ 'r', 1, "read", "n", "Read n bytes" },
	{ 's', 2, "signed", "m n", "Get the next m * n bytes as m byte signed integers" },
	{ 'u', 2, "unsigned", "m n", "Get the next m * n bytes as m byte unsigned integers" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'z', 1, "readz", "n", "Read n bytes and fill with zero" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"readbin: read binary files",
		"usage: readbin [options]",
		opts
	);

	fflush (stdout);
}

static
void prt_uint32 (FILE *out, unsigned long val)
{
	unsigned i;
	unsigned dig;
	char     buf[256];

	i = 0;

	do {
		dig = (unsigned) (val % par_base);
		val = val / par_base;

		if (dig < 10 ) {
			buf[i] = (char) ('0' + dig);
		}
		else {
			buf[i] = (char) ('A' + dig - 10);
		}

		i += 1;
	} while ((val != 0) && (i < 256));

	while (i > 0) {
		i -= 1;
		fputc (buf[i], out);
	}

	fputc ('\n', out);

	fflush (out);
}

static
void prt_sint32 (FILE *out, long val)
{
	if (val < 0) {
		val = -val;
		fputc ('-', out);
	}

	prt_uint32 (out, (unsigned long) val);
}

static
unsigned get_uint8 (unsigned char c[1])
{
	return (c[0]);
}

static
unsigned get_uint16 (unsigned char c[2])
{
	unsigned r;

	if (par_little) {
		r = c[1] & 0xff;
		r = (r << 8) + (c[0] & 0xff);
	}
	else {
		r = c[0] & 0xff;
		r = (r << 8) + (c[1] & 0xff);
	}

	return (r);
}

static
unsigned long get_uint32 (unsigned char c[4])
{
	unsigned long r;

	if (par_little) {
		r = c[3] & 0xff;
		r = (r << 8) + (c[2] & 0xff);
		r = (r << 8) + (c[1] & 0xff);
		r = (r << 8) + (c[0] & 0xff);
	}
	else {
		r = c[0] & 0xff;
		r = (r << 8) + (c[1] & 0xff);
		r = (r << 8) + (c[2] & 0xff);
		r = (r << 8) + (c[3] & 0xff);
	}

	return (r);
}

static
int get_sint8 (unsigned char c[1])
{
	int r;

	r = c[0] & 0xff;

	if (r > 127) {
		r = r - 256;
	}

	return (r);
}

static
int get_sint16 (unsigned char c[2])
{
	long     r;

	r = (long) get_uint16 (c);

	if (r > 32767) {
		r = r - 65536;
	}

	return ((int) r);
}

static
long get_sint32 (unsigned char c[4])
{
	long          r;
	unsigned long tmp;

	tmp = get_uint32 (c);

	if (tmp >= (1 << 31)) {
		r = (long) (tmp - (1 << 31));
		r = r - (1 << 31);
	}
	else {
		r = (long) tmp;
	}

	return (r);
}

static
int rb_bytes (FILE *inp, FILE *out, unsigned long n, int fill)
{
	unsigned long i;
	unsigned char buf[MISC_READBIN_BUFFER];

	while (n > 0) {
		i = (n < MISC_READBIN_BUFFER) ? n : MISC_READBIN_BUFFER;

		i = fread (buf, 1, i, inp);
		if (i == 0) {
			break;
		}

		if (fwrite (buf, 1, i, out) != i) {
			return (1);
		}

		n -= i;
	}

	if (n == 0) {
		return (0);
	}

	if (fill == 0) {
		return (1);
	}

	memset (buf, 0, MISC_READBIN_BUFFER);

	while (n > 0) {
		i = (n < MISC_READBIN_BUFFER) ? n : MISC_READBIN_BUFFER;

		if (fwrite (buf, 1, i, out) != i) {
			return (1);
		}

		n -= i;
	}

	return (0);
}

static
int rb_skip (FILE *inp, FILE *out, unsigned long n)
{
	unsigned long i;
	unsigned char buf[MISC_READBIN_BUFFER];

	if (n <= 0x7fffffffUL) {
		if (fseek (inp, (long) n, SEEK_CUR) == 0) {
			return (0);
		}
	}

	while (n > 0) {
		i = (n < MISC_READBIN_BUFFER) ? n : MISC_READBIN_BUFFER;

		i = fread (buf, 1, i, inp);
		if (i == 0) {
			return (0);
		}

		n -= i;
	}

	return (0);
}

static
int rb_pos (FILE *inp, FILE *out, unsigned long n)
{
	if (n <= 0x7fffffffUL) {
		if (fseek (inp, (long) n, SEEK_SET) == 0) {
			return (0);
		}
	}

	return (1);
}

static
int rb_cat (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned char buf[MISC_READBIN_BUFFER];

	n = fread (buf, 1, MISC_READBIN_BUFFER, inp);

	while (n > 0) {
		if (fwrite (buf, 1, n, out) != n) {
			return (1);
		}

		n = fread (buf, 1, MISC_READBIN_BUFFER, inp);
	}

	return (0);
}

static
int rb_int8 (FILE *inp, FILE *out, int sign, unsigned long n)
{
	unsigned long i;
	unsigned char buf[1];

	for (i = 0; i < n; i++) {
		if (fread (buf, 1, 1, inp) != 1) {
			return (1);
		}

		if (sign) {
			prt_sint32 (out, get_sint8 (buf));
		}
		else {
			prt_uint32 (out, get_uint8 (buf));
		}
	}

	return (0);
}

static
int rb_int16 (FILE *inp, FILE *out, int sign, unsigned long n)
{
	unsigned long i;
	unsigned char buf[2];

	for (i = 0; i < n; i++) {
		if (fread (buf, 1, 2, inp) != 2) {
			return (1);
		}

		if (sign) {
			prt_sint32 (out, get_sint16 (buf));
		}
		else {
			prt_uint32 (out, get_uint16 (buf));
		}
	}

	return (0);
}

static
int rb_int32 (FILE *inp, FILE *out, int sign, unsigned long n)
{
	unsigned long i;
	unsigned char buf[4];

	for (i = 0; i < n; i++) {
		if (fread (buf, 1, 4, inp) != 4) {
			return (1);
		}

		if (sign) {
			prt_sint32 (out, get_sint32 (buf));
		}
		else {
			prt_uint32 (out, get_uint32 (buf));
		}
	}

	return (0);
}

static
int rb_unsigned (FILE *inp, FILE *out, unsigned long m, unsigned long n)
{
	switch (m) {
	case 0:
		return (1);

	case 1:
		return (rb_int8 (inp, out, 0, n));

	case 2:
		return (rb_int16 (inp, out, 0, n));

	case 4:
		return (rb_int32 (inp, out, 0, n));

	default:
		return (1);
	}

	return (1);
}

static
int rb_signed (FILE *inp, FILE *out, unsigned long m, unsigned long n)
{
	switch (m) {
	case 0:
		return (1);

	case 1:
		return (rb_int8 (inp, out, 1, n));

	case 2:
		return (rb_int16 (inp, out, 1, n));

	case 4:
		return (rb_int32 (inp, out, 1, n));

	default:
		return (1);
	}

	return (1);
}

int main (int argc, char *argv[])
{
	int           r;
	char          **optarg;
	FILE          *inp, *out;
	unsigned long m, n;

	argv0 = argv[0];

	inp = misc_open_inp (argv0, NULL, "rb");
	out = misc_open_out (argv0, NULL, "wb");

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("readbin");
			return (EXIT_SUCCESS);

		case 'b':
			par_little = 0;
			break;

		case 'c':
			if (rb_cat (inp, out)) {
				return (EXIT_FAILURE);
			}
			break;

		case 'i':
			misc_close (inp);
			inp = misc_open_inp (argv0, optarg[0], "rb");

			if (inp == NULL) {
				return (EXIT_FAILURE);
			}
			break;

		case 'k':
			if (str_get_ulng (optarg[0], &n)) {
				fprintf (stderr, "%s: bad number (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			if (rb_skip (inp, out, n)) {
				return (EXIT_FAILURE);
			}
			break;

		case 'l':
			par_little = 1;
			break;

		case 'n':
			if (str_get_uint (optarg[0], &par_base)) {
				fprintf (stderr, "%s: bad base (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			misc_close (out);
			out = fopen (optarg[0], "wb");

			if (out == NULL) {
				return (EXIT_FAILURE);
			}
			break;

		case 'p':
			if (str_get_ulng (optarg[0], &n)) {
				fprintf (stderr, "%s: bad number (%s)\n",
					argv0, optarg[0]
				);
				return (1);
			}

			if (rb_pos (inp, out, n)) {
				return (EXIT_FAILURE);
			}
			break;

		case 'r':
			if (str_get_ulng (optarg[0], &n)) {
				fprintf (stderr, "%s: bad number (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			if (rb_bytes (inp, out, n, 0)) {
				return (EXIT_FAILURE);
			}
			break;

		case 's':
			if (str_get_ulng (optarg[0], &m)) {
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (optarg[1], &n)) {
				return (EXIT_FAILURE);
			}

			if (rb_signed (inp, out, m, n)) {
				return (EXIT_FAILURE);
			}
			break;

		case 'u':
			if (str_get_ulng (optarg[0], &m)) {
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (optarg[1], &n)) {
				return (EXIT_FAILURE);
			}

			if (rb_unsigned (inp, out, m, n)) {
				return (EXIT_FAILURE);
			}
			break;

		case 'z':
			if (str_get_ulng (optarg[0], &n)) {
				fprintf (stderr, "%s: bad number (%s)\n",
					argv0, optarg[0]
				);
				return (1);
			}

			if (rb_bytes (inp, out, n, 1)) {
				return (EXIT_FAILURE);
			}
			break;

		case 0:
			misc_close (inp);
			inp = misc_open_inp (argv0, optarg[0], "rb");

			if (inp == NULL) {
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	misc_close (out);
	misc_close (inp);

	return (EXIT_SUCCESS);
}
