/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/sched/sched.c                                            *
 * Created:     2002-03-21 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/strtoint.h>

#include <unistd.h>
#include <sched.h>

#include "sched.h"


static const char *argv0 = NULL;

static int   par_verbose = 0;
static char  *par_cmd = NULL;
static char  **par_arg = NULL;

static int   par_policy = SCHED_OTHER;
static char  *par_policy_name = "SCHED_OTHER";
static int   par_priority = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'c', 1, "command", "string", "Set command to execute" },
	{ 'n', 1, "priority", "int", "Set the static priority [0]" },
	{ 'p', 1, "policy", "string", "Set scheduling policy [other]" },
	{ 'q', 0, "query", NULL, "Print the current scheduling policy and exit" },
	{ 'v', 0, "verbose", NULL, "Verbose operation [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"sched: set scheduling policy",
		"usage: sched [options]",
		opts
	);

	fputs (
		"\n"
		"policies are:\n"
		"  other\n"
		"  fifo\n"
		"  rr\n",
		stdout
	);

	fflush (stdout);
}

static
int do_query (void)
{
	int                pol;
	int                p_max, p_min;
	char               *str;
	struct sched_param p;

	pol = sched_getscheduler (0);

	if (sched_getparam (0, &p)) {
		fprintf (stderr, "%s: could not get schduling parameters\n",
			argv0
		);
		return (1);
	}

	p_min = sched_get_priority_min (pol);
	p_max = sched_get_priority_max (pol);

	switch (pol) {
	case SCHED_OTHER:
		str = "SCHED_OTHER";
		break;

	case SCHED_FIFO:
		str = "SCHED_FIFO";
		break;

	case SCHED_RR:
		str = "SCHED_RR";
		break;

	default:
		str = "unknown";
		break;
	};

	fprintf (stdout,
		"policy: %s (%d)  priority: %d  min: %d  max: %d\n",
		str, pol, p.sched_priority, p_min, p_max
	);

	return (0);
}

static
int sched_run_cmd (void)
{
	struct sched_param p;

	if (sched_getparam (0, &p)) {
		fprintf (stderr, "%s: could not get schduling parameters\n",
			argv0
		);
		return (1);
	}

	p.sched_priority = par_priority;

	if (sched_setscheduler (0, par_policy, &p)) {
		fprintf (stderr, "%s: can't set scheduling policy\n", argv0);
		return (1);
	}

	execvp (par_cmd, par_arg);

	fprintf (stderr, "%s: can't exec command (%s)\n", argv0, par_cmd);

	return (1);
}

static
int sched_set_policy (const char *str)
{
	if (strcmp (str, "other") == 0) {
		par_policy = SCHED_OTHER;
		par_policy_name = "SCHED_OTHER";
	}
	else if (strcmp (str, "fifo") == 0) {
		par_policy = SCHED_FIFO;
		par_policy_name = "SCHED_FIFO";
	}
	else if (strcmp (str, "rr") == 0) {
		par_policy = SCHED_RR;
		par_policy_name = "SCHED_RR";
	}
	else {
		if (str_get_sint (str, &par_policy)) {
			return (1);
		}

		par_policy_name = "numeric";
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;

	argv0 = argv[0];

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		if ((r == 'c') || (r == 0)) {
			par_cmd = optarg[0];
			par_arg = optarg;

			break;
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("sched");
			return (EXIT_SUCCESS);

		case 'n':
			if (strcmp (optarg[0], "min")) {
				par_priority = sched_get_priority_min (par_policy);
			}
			else if (strcmp (optarg[0], "max")) {
				par_priority = sched_get_priority_max (par_policy);
			}
			else {
				if (str_get_sint (optarg[0], &par_priority)) {
					fprintf (stderr,
						"%s: bad priority (%s)\n",
						argv0, optarg[0]
					);
					return (EXIT_FAILURE);
				}
			}
			break;

		case 'p':
			if (sched_set_policy (optarg[0])) {
				fprintf (stderr,
					"%s: bad scheduling policy (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'q':
			if (do_query()) {
				return (EXIT_FAILURE);
			}
			return (EXIT_SUCCESS);

		case 'v':
			par_verbose = 1;
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_verbose) {
		fprintf (stderr, "policy:   %s (%d)\n", par_policy_name, par_policy);
		fprintf (stderr, "priority: %d\n", par_priority);
	}

	if (par_cmd == NULL) {
		return (EXIT_SUCCESS);
	}

	r = sched_run_cmd();

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
