/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/sparse/sparse.c                                          *
 * Created:     2003-07-09 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2003-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/config.h>
#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include <stdlib.h>

#ifdef HAVE_FTRUNCATE
#include <unistd.h>
#include <sys/types.h>
#endif

#include "sparse.h"


#define SPARSE_BUF_SIZE MISC_BUF_SIZE


static const char *argv0 = NULL;

#ifdef HAVE_FSEEKO
static off_t par_zero_cnt = 0;
static off_t par_size = 0;
#else
static long par_zero_cnt = 0;
static long par_size = 0;
#endif

static int   par_create = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'c', 1, "create", "size", "Create an empty sparse file" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"sparse: create sparse files",
		"usage: sparse [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
int sparse_write (FILE *fp, const void *buf, unsigned cnt)
{
	unsigned            i;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	i = 0;
	while ((i < cnt) && (tmp[i] == 0)) {
		par_zero_cnt += 1;
		i += 1;
	}

	if (i >= cnt) {
		return (0);
	}

	if (par_zero_cnt > 0) {
#ifdef HAVE_FSEEKO
		if (fseeko (fp, par_zero_cnt, SEEK_CUR)) {
			return (1);
		}
#else
		if (fseek (fp, par_zero_cnt, SEEK_CUR)) {
			return (1);
		}
#endif

		par_zero_cnt = 0;
	}

	if (fwrite (tmp + i, cnt - i, 1, fp) != 1) {
		return (1);
	}

	return (0);
}

static
int sparse_flush (FILE *fp)
{
	if (par_zero_cnt == 0) {
		return (0);
	}

#ifdef HAVE_FSEEKO
	if (fseeko (fp, par_zero_cnt - 1, SEEK_CUR)) {
		return (1);
	}
#else
	if (fseek (fp, par_zero_cnt - 1, SEEK_CUR)) {
		return (1);
	}
#endif

	par_zero_cnt = 0;

	fputc (0, fp);

	fflush (fp);

	return (0);
}

static
int sparse_make (FILE *inp, FILE *out)
{
	size_t        r;
	unsigned char buf[SPARSE_BUF_SIZE];

	while (1) {
		r = fread (buf, 1, SPARSE_BUF_SIZE, inp);

		if (r < 0) {
			return (1);
		}

		if (r == 0) {
			break;
		}

		if (sparse_write (out, buf, r)) {
			return (1);
		}

		par_size += r;
	}

	if (sparse_flush (out)) {
		return (1);
	}

	return (0);
}

#ifdef HAVE_FTRUNCATE
static
int sparse_create (FILE *fp)
{
	int fd;

	fd = fileno (fp);

	if (ftruncate (fd, par_size)) {
		return (1);
	}

	return (0);
}
#else
static
int sparse_create (FILE *fp)
{
	par_zero_cnt = par_size;

	if (sparse_flush (fp)) {
		return (1);
	}

	return (0);
}
#endif

static
int str_get_ullng (const char *str, unsigned long long *ret)
{
	unsigned long long val;
	char               *end;

	val = strtoull (str, &end, 0);

	if (end == str) {
		return (1);
	}

	if ((*end == 'k') || (*end == 'K')) {
		val *= 1024;
	}
	else if ((*end == 'm') || (*end == 'M')) {
		val *= 1024;
		val *= 1024;
	}
	else if ((*end == 'g') || (*end == 'G')) {
		val *= 1024;
		val *= 1024;
		val *= 1024;
	}

	*ret = val;

	return (0);
}

int main (int argc, char *argv[])
{
	int                r;
	char               **optarg;
	const char         *finp, *fout;
	FILE               *inp, *out;
	unsigned long long tmp;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("sparse");
			return (EXIT_SUCCESS);

		case 'c':
			if (str_get_ullng (optarg[0], &tmp)) {
				fprintf (stderr, "%s: bad file size (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			par_size = (off_t) tmp;

			if ((par_size < 0) || ((unsigned long long) par_size != tmp)) {
				fprintf (stderr,
					"%s: file size too large (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			par_create = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, "rb");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, "wb");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	if (par_create) {
		r = sparse_create (out);
	}
	else {
		r = sparse_make (inp, out);
	}

	misc_close (out);
	misc_close (inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
